from abc import ABC, abstractclassmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import date
from typing import Dict, List

from lazy_budget.budget import Budget, FinancialOperation
from lazy_budget.money_provider import Currency, Money


@dataclass
class BaseBudgetStats(ABC):
    @abstractclassmethod
    def get_stats(
        cls, budget: Budget, operations: List[FinancialOperation]
    ) -> "BaseBudgetStats":
        raise NotImplementedError


@dataclass
class BasicBudgetStats(BaseBudgetStats):
    today: date
    total_days: int
    days_left: int
    available_per_day: Money
    can_spend_today: Money
    avg_spent_per_day: Money
    currently_keeping: Money
    spent_today: Money
    currency: Currency
    days_until_can_spend: int = 0
    budget: Budget = field(compare=False, hash=False, repr=False, default=None)

    @classmethod
    def get_stats(
        cls, budget: Budget, operations: List[FinancialOperation]
    ) -> "BasicBudgetStats":
        if not budget.is_ongoing:
            raise ValueError("there can be no stats for a finished budget")

        operations = budget.filter(operations)
        can_spend_today = budget.get_can_spend_today(operations)
        available_per_day = budget.available_per_day
        return cls(
            today=budget.today,
            total_days=budget.total_days,
            days_left=budget.days_left,
            available_per_day=available_per_day,
            can_spend_today=can_spend_today,
            days_until_can_spend=budget.get_days_until_can_spend(
                can_spend_today, available_per_day
            ),
            avg_spent_per_day=budget.get_avg_spent_per_day(operations),
            currently_keeping=budget.get_currently_keeping(
                can_spend_today=can_spend_today
            ),
            spent_today=budget.get_spent_today(operations),
            currency=budget.currency,
            budget=budget,
        )


def get_categories_to_spendings(
    budget: Budget, operations: List[FinancialOperation]
) -> Dict[str, Money]:
    categories_to_spendings = defaultdict(lambda: budget.zero)
    for operation in operations:
        categories_to_spendings[operation.category] += operation.money_value

    return dict(sorted(categories_to_spendings.items(), key=lambda x: x[1]))



from abc import abstractmethod
from dataclasses import dataclass
from itertools import takewhile
from typing import Iterable

import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from pandas.core.frame import DataFrame

from lazy_budget.budget import Budget, FinancialOperation
from lazy_budget.stats import get_categories_to_spendings


@dataclass
class BaseChart:
    budget: Budget
    operations: Iterable[FinancialOperation]

    def plot(self):
        df = self.get_dataframe()
        self.display_dataframe(df)
        plt.show()

    @abstractmethod
    def get_dataframe(self) -> pd.DataFrame:
        pass

    @abstractmethod
    def display_dataframe(self, dataframe: pd.DataFrame):
        pass


@dataclass
class BurnDownChart(BaseChart):
    show_zero: bool = False

    def get_dataframe(self):
        max_date = self.budget.today
        operations = self.budget.filter(self.operations)
        spent_per_day = self.budget.get_spent_by_day(operations)
        for_df = []
        total_spent = self.budget.total.amount
        for day in takewhile(lambda x: x <= max_date, iter(self.budget)):
            self.budget.today = day
            ops = self.budget.filter(operations)
            total_spent += spent_per_day[day].amount
            should_be_available = (
                self.budget.days_left * self.budget.available_per_day.amount
                + self.budget.keep.amount
            )
            for_df.append(
                (
                    day.strftime("%d"),
                    total_spent,
                    should_be_available,
                    0,
                    self.budget.available_per_day.amount,
                    self.budget.get_currently_keeping(ops).amount,
                    self.budget.keep.amount,
                )
            )
        return pd.DataFrame(
            for_df,
            columns=[
                "day",
                "money",
                "should_be_available",
                "zero",
                "available_per_day",
                "currently_keeping",
                "keep",
            ],
        )

    def display_dataframe(self, dataframe: DataFrame):
        sns.set_theme()
        sns.set_style("darkgrid")

        plt.plot("day", "money", data=dataframe, color="red", label="real burndown")
        plt.plot(
            "day",
            "should_be_available",
            data=dataframe,
            linestyle="dashed",
            color="red",
            label="planned burndown",
        )
        if self.show_zero:
            plt.plot("day", "zero", data=dataframe, linestyle="dashed")

        plt.plot(
            "day",
            "currently_keeping",
            data=dataframe,
            color="green",
            label="real savings",
        )
        plt.plot(
            "day",
            "keep",
            data=dataframe,
            color="green",
            linestyle=":",
            label="planned savings",
        )
        plt.legend()


class CategoryBarPlot(BaseChart):
    def get_dataframe(self) -> pd.DataFrame:
        operations = self.budget.filter(self.operations)
        sps = get_categories_to_spendings(self.budget, operations)
        return pd.DataFrame(
            map(lambda x: (x[0], x[1].amount), sps.items()),
            columns=["category", "money_spent"],
        )

    def display_dataframe(self, dataframe: pd.DataFrame):
        sns.set(style="darkgrid")

        # Set the figure size
        # plt.figure(figsize=(10, 7))

        # plot a bar chart
        ax = sns.barplot(
            x="category",
            y="money_spent",
            data=dataframe,
            estimator=sum,
            ci=None,
            color="#69b3a2",
        )
        ax.set(xlabel="category", ylabel=f"money spent, {self.budget.currency.name}")
        plt.gca().invert_yaxis()



from locale import getdefaultlocale

DAYS_IN_A_SECOND = 0.0000115741
BUDGET_FILE = "./budget.yml"
HISTORY_FILE = "./history.csv"
DEFAULT_LOCALE = getdefaultlocale()[0]
CONFIG_FILE_PATH = "~/.lbudrc.yml"
